#include <iostream>
#include <vector>
#include <map>
#include <utility>
#include <numeric>
#include <limits>
#include <cmath>
#include <TSystem.h>

#include "filval/filval.hpp"
#include "filval/root/filval.hpp"

#include <boost/range/combine.hpp>
#include <boost/format.hpp>

#include "analysis/TrackingNtuple.h"
#include "analysis/TrackingNtupleObjs.hpp"
#include "analysis/common.hpp"

using namespace std;
using namespace fv;
using namespace fv::root;

typedef std::pair<std::vector<float>,std::vector<float>> pair_vec;

#define HIT_TYPE_PIXEL 0
#define HIT_TYPE_GLUED 1
#define HIT_TYPE_STRIP 2

enum PixSubDet{
    Void = 0,
    Barrel = 1,
    EndCap = 2
};

struct PixDetElem{
    PixSubDet subdet;
    unsigned char layer;

    PixDetElem():subdet(PixSubDet::Void), layer(0) {}

    PixDetElem(const PixSubDet& subdet, const unsigned char& layer){
        this->subdet = subdet;
        this->layer = layer;
    }

    template<typename T>
    bool contains(const T& obj) const{
        return obj.subdet() == subdet && obj.layer() == layer;
    }

    bool contains(const PixSubDet& subdet, const unsigned char& layer) const{
        return subdet == this->subdet && layer == this->layer;
    }
};

std::map<PixSubDet,string> subdet_names = {{PixSubDet::Barrel, "BPIX"}, {PixSubDet::EndCap, "FPIX"}};


typedef std::tuple<PixRecHit, SimHit> HitPair;

vector<string> seedTypes =
    {"initialStepSeeds",
     "highPtTripletStepSeeds",
     "mixedTripletStepSeeds",
     "pixelLessStepSeeds",
     "tripletElectronSeeds",
     "pixelPairElectronSeeds",
     "stripPairElectronSeeds"};

SeedCollection seeds;
SimTrackCollection sim_tracks;
SimVertexCollection sim_vertices;

void register_objects(TrackingDataSet& tds){
    seeds.init(tds);
    sim_tracks.init(tds);
    sim_vertices.init(tds);
}

float rad(const float& x, const float& y) {
    return sqrt(x*x + y*y);
}

void run_analysis(const vector<fv::util::DataFileDescriptor>& dfds, const string output_filename, bool silent){
    gSystem->Load("libfilval.so");
    auto replace_suffix = [](const std::string& input, const std::string& new_suffix){
        return input.substr(0, input.find_last_of(".")) + new_suffix;
    };

    string log_filename = replace_suffix(output_filename, ".log");
    fv::util::Log::init_logger(log_filename, fv::util::LogPriority::kLogInfo);

    TrackingDataSet tds(output_filename, dfds, "trackingNtuple/tree");
    /* tds.set_max_events(11); */
    register_objects(tds);

    // Indices of prompt gen electrons
    auto prompt_electrons = func_value<vector<size_t>>("prompt_electrons",
        FUNC(([](){
            vector<size_t> idxs;
            for(const auto& sim_track : sim_tracks) {
                if (abs(sim_track.pdgId()) != 11) continue;
                const SimVertex& sim_vertex = sim_vertices[sim_track.parentVtxIdx()];
                float rho = rad(sim_vertex.x(), sim_vertex.y());
                if (rho > 1) continue;
                idxs.push_back(sim_track.idx);
            }
            return idxs;
        })));

    // Indices of prompt gen electrons that have a matched seed
    function<bool(size_t)> with_seed = [](const size_t& e_idx) {
        if (seeds.size() == 0 or
            sim_tracks[e_idx].seedIdx().size() == 0) return false;
        else return true;
    };

    // Indices of prompt gen electrons that have a matched gsftrack
    function<bool(size_t)> with_track = [](const size_t& e_idx) {
        return sim_tracks[e_idx].trkIdx().size() > 0;
    };

    function<float(size_t)> e_pt = [](const size_t& e_idx) {
        return sim_tracks[e_idx].pt();
    };

    function<float(size_t)> e_eta = [](const size_t& e_idx) {
        return sim_tracks[e_idx].eta();
    };

    function<float(size_t)> e_phi = [](const size_t& e_idx) {
        return sim_tracks[e_idx].phi();
    };

    tds.register_container<EfficiencyContainer<size_t>>("seed_eff_v_pt",
        prompt_electrons, TH1Params::lookup("eff_v_pt"), nullptr, &with_seed, &e_pt);

    tds.register_container<EfficiencyContainer<size_t>>("seed_eff_v_eta",
        prompt_electrons, TH1Params::lookup("eff_v_eta"), nullptr, &with_seed, &e_eta);

    tds.register_container<EfficiencyContainer<size_t>>("seed_eff_v_phi",
        prompt_electrons, TH1Params::lookup("eff_v_phi"), nullptr, &with_seed, &e_phi);

    tds.register_container<EfficiencyContainer<size_t>>("track_eff_v_pt",
        prompt_electrons, TH1Params::lookup("eff_v_pt"), nullptr, &with_track, &e_pt);

    tds.register_container<EfficiencyContainer<size_t>>("track_eff_v_eta",
        prompt_electrons, TH1Params::lookup("eff_v_eta"), nullptr, &with_track, &e_eta);

    tds.register_container<EfficiencyContainer<size_t>>("track_eff_v_phi",
        prompt_electrons, TH1Params::lookup("eff_v_phi"), nullptr, &with_track, &e_phi);

    /* std::ofstream printout("track_eff.txt"); */
    /* tds.register_container<Printer<vector<size_t>>>("prompt_electron_idx", prompt_electrons, &printout); */
    /* tds.register_container<Printer<vector<size_t>>>("with track:        ", apply(with_seed, prompt_electrons), &printout); */

    /* tds.register_container<ContainerTH1Many<float>>("dxys", sim_tracks.val_pca_dxy, "dxys", */
    /*                                                 TH1Params::lookup("dxys")); */

    tds.process(silent);
    tds.save_all();
}

int main(int argc, char * argv[]){
    fv::util::ArgParser args(argc, argv);
    bool silent = args.cmdOptionExists("-s");
    string output_filename = args.cmdOptionExists("-o") ? args.getCmdOption("-o") : "output.root";
    if(args.cmdOptionExists("-c")){
        fv::util::init_config(args.getCmdOption("-c"));
    }
    if(args.cmdOptionExists("-F")){
        auto file_list = fv::util::read_input_list(args.getCmdOption("-F"));
        run_analysis(file_list, output_filename, silent);
    }else if(args.cmdOptionExists("-f")){
        string input_filename = args.getCmdOption("-f");
        string data_label = args.cmdOptionExists("-l") ? args.getCmdOption("-l") : "";
        string data_category = args.cmdOptionExists("-c") ? args.getCmdOption("-c") : "";
        fv::util::DataFileDescriptor dfd(input_filename, data_label, data_category);
        run_analysis(std::vector<fv::util::DataFileDescriptor>({dfd}), output_filename, silent);
    }else {
        cout << "Usage: ./tracking_eff (-s) {-o output_filename} -F datafiles.txt" << endl;
        cout << "       ./tracking_eff (-s) {-l DATA_LABEL} {-c DATA_CATEGORY} {-o output_filename} -f treefile.root" << endl;
    }
    return 0;
}
